from typing import Union, Tuple, Sequence

from neuralogic.core.constructs.relation import BaseRelation
from neuralogic.core.constructs.function import Transformation, Combination, Function, Aggregation


dot_type = type(Ellipsis)


# Transformation


def sigmoid(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.SIGMOID(entity)


def tanh(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.TANH(entity)


def signum(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.SIGNUM(entity)


def relu(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.RELU(entity)


def leaky_relu(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.LEAKY_RELU(entity)


def lukasiewicz(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.LUKASIEWICZ(entity)


def exp(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.EXP(entity)


def sqrt(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.SQRT(entity)


def inverse(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.INVERSE(entity)


def reverse(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.REVERSE(entity)


def log(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.LOG(entity)


def identity(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.IDENTITY(entity)


def transp(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.TRANSP(entity)


def softmax(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.SOFTMAX(entity)


def sparsemax(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.SPARSEMAX(entity)


def norm(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Transformation.NORM(entity)


def slice(
    entity: Union[BaseRelation, Function] = None,
    *,
    rows: Union[type(Ellipsis), Tuple[int, int]] = ...,
    cols: Union[type(Ellipsis), Tuple[int, int]] = ...,
) -> Union[BaseRelation, Function]:
    r"""
    Slices a value into a new value that is created by taking values on specified rows and columns.

    Rows and Cols coordinates are specified either as ``...``, which means all rows/cols or by a tuple of two
    elements ``[from_index, to_index]``.

    Parameters
    ----------

    entity : Union[BaseRelation, Function]
        Relation to apply the function on. Default: ``None``
    rows : Union[type(Ellipsis), Tuple[int, int]]
        Default: ``...``
    cols : Union[type(Ellipsis), Tuple[int, int]]
        Default: ``...``
    """
    return Transformation.SLICE(entity, rows=rows, cols=cols)


def reshape(
    entity: Union[BaseRelation, Function] = None,
    *,
    shape: Union[None, Tuple[int, int], int],
) -> Union[BaseRelation, Function]:
    r"""
    Change the shape/type of the value to a new shape. The shape can be either ``None``, int, or a tuple of two ints.

    * If ``None``, the underlying value will be converted to a scalar. E.g., a matrix value of one element ``[[1]]``
      will be converted to scalar ``1``.

    * If int, then the value will be converted to scalar (if the int is ``0``) or to a column vector.

    * If a tuple of two ints, the value will be converted to a scalar if the tuple is ``(0, 0)``. Into a row vector
      if the shape is ``(len, 0)`` or a column vector for shape ``(0, len)``. For other tuples ``(n, m)``,
      the value will be reshaped to matrix :math:`n \times m`.

    Parameters
    ----------

    entity : Union[BaseRelation, Function]
        Relation to apply the function on. Default: ``None``
    shape : Union[None, Tuple[int, int], int]
        The new shape of the value
    """
    return Transformation.RESHAPE(entity, shape=shape)


# Combination


def max_comb(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Combination.MAX(entity)


def min_comb(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Combination.MIN(entity)


def avg_comb(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Combination.AVG(entity)


def sum_comb(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Combination.SUM(entity)


def count_comb(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Combination.COUNT(entity)


def product_comb(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Combination.PRODUCT(entity)


def elproduct_comb(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Combination.ELPRODUCT(entity)


def softmax_comb(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Combination.SOFTMAX(entity)


def sparsemax_comb(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Combination.SPARSEMAX(entity)


def crosssum_comb(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Combination.CROSSSUM(entity)


def concat_comb(entity: Union[BaseRelation, Function] = None, *, axis: int = -1) -> Union[BaseRelation, Function]:
    return Combination.CONCAT(entity, axis=axis)


def cossim_comb(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Combination.COSSIM(entity)


# Aggregations


def max(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Aggregation.MAX(entity)


def min(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Aggregation.MIN(entity)


def avg(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Aggregation.AVG(entity)


def sum(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Aggregation.SUM(entity)


def count(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Aggregation.COUNT(entity)


def concat(entity: BaseRelation = None, *, axis: int = -1) -> Union[BaseRelation, Function]:
    return Aggregation.CONCAT(entity, axis=axis)


def softmax_agg(entity: BaseRelation = None, *, agg_terms: Sequence[str] = None) -> Union[BaseRelation, Function]:
    return Aggregation.SOFTMAX(entity, agg_terms=agg_terms)



from typing import Dict, Optional, Union, Callable, List

from neuralogic.core.settings import Settings
from neuralogic.core.builder import DatasetBuilder
from neuralogic.core import Template, BuiltDataset, SettingsProxy
from neuralogic.dataset.base import BaseDataset

from neuralogic.utils.visualize import draw_model


class AbstractNeuraLogic:
    def __init__(self, dataset_builder: DatasetBuilder, template: Template, settings: SettingsProxy):
        self.need_sync = True

        self.source_template = [rule for rule in template.template]
        self.template = dataset_builder.parsed_template
        self.dataset_builder = dataset_builder

        self.settings = settings

        self.hooks_set = False
        self.hooks: Dict[str, List[Callable]] = {}

    def __call__(self, sample):
        raise NotImplementedError

    def build_dataset(
        self,
        dataset: BaseDataset,
        *,
        batch_size: int = 1,
        file_mode: bool = False,
        learnable_facts: bool = False,
        progress: bool = False,
    ):
        return self.dataset_builder.build_dataset(
            dataset,
            self.settings,
            batch_size=batch_size,
            file_mode=file_mode,
            learnable_facts=learnable_facts,
            progress=progress,
        )

    def set_hooks(self, hooks):
        self.hooks_set = len(hooks) != 0
        self.hooks = hooks

    def run_hook(self, hook: str, value):
        for callback in self.hooks[hook]:
            callback(value)

    def sync_template(self, state_dict: Optional[Dict] = None, weights=None):
        state_dict = self.state_dict() if state_dict is None else state_dict
        weights = self.template.getAllWeights() if weights is None else weights
        weight_dict = state_dict["weights"]

        for weight in weights:
            if not weight.isLearnable:
                continue
            weight_value = weight.value

            index = weight.index
            value = weight_dict[index]

            if isinstance(value, (float, int)):
                weight_value.set(0, float(value))
                continue

            if isinstance(value[0], (float, int)):
                for i, val in enumerate(value):
                    weight_value.set(i, float(val))
                continue

            cols = len(value[0])

            for i, values in enumerate(value):
                for j, val in enumerate(values):
                    weight_value.set(i * cols + j, float(val))

    def parameters(self) -> Dict:
        return self.state_dict()

    def state_dict(self) -> Dict:
        raise NotImplementedError

    def load_state_dict(self, state_dict: Dict):
        raise NotImplementedError

    def draw(
        self,
        filename: Optional[str] = None,
        draw_ipython=True,
        img_type="png",
        value_detail: int = 0,
        graphviz_path: Optional[str] = None,
        *args,
        **kwargs,
    ):
        return draw_model(self, filename, draw_ipython, img_type, value_detail, graphviz_path, *args, **kwargs)


class AbstractEvaluator:
    def __init__(self, template: Template, settings: Settings):
        self.settings = settings.create_proxy()

        self.neuralogic_model = template.build(settings)
        self.neuralogic_model.set_hooks(template.hooks)

    def build_dataset(
        self,
        dataset: Union[BaseDataset, BuiltDataset],
        *,
        batch_size: int = 1,
        file_mode: bool = False,
        learnable_facts: bool = False,
        progress: bool = False,
    ):
        if isinstance(dataset, BaseDataset):
            return self.neuralogic_model.build_dataset(
                dataset,
                batch_size=batch_size,
                file_mode=file_mode,
                learnable_facts=learnable_facts,
                progress=progress,
            )
        return dataset

    @property
    def model(self) -> AbstractNeuraLogic:
        return self.neuralogic_model

    def train(self, dataset: Optional[Union[BaseDataset, BuiltDataset]] = None, *, generator: bool = True):
        pass

    def test(self, dataset: Optional[Union[BaseDataset, BuiltDataset]] = None, *, generator: bool = True):
        pass

    def parameters(self) -> Dict:
        return self.state_dict()

    def state_dict(self) -> Dict:
        pass

    def load_state_dict(self, state_dict: Dict):
        pass

    def reset_parameters(self):
        self.neuralogic_model.reset_parameters()

    def draw(
        self,
        filename: Optional[str] = None,
        draw_ipython=True,
        img_type="png",
        value_detail: int = 0,
        graphviz_path: Optional[str] = None,
        *args,
        **kwargs,
    ):
        return self.neuralogic_model.draw(
            filename, draw_ipython, img_type, value_detail, graphviz_path, *args, **kwargs
        )
